;*********************************************************************************************************
;		Zyklengenaue Emulation eines MOS6510 auf einem ATmega
;			(c) Mark Leyer 2010
;*********************************************************************************************************




;*********************************************************************************************************
;	Variable Festlegungen fr den Emulator:

.include "m644def.inc"


.equ Vector6510Reset 	= 0x0200			; Startadresse fr den 6510
.equ Vector6510IRQ		= 0x0200
.equ Vector6510NMI		= 0x0200

; Port B fr RDY, NMI und IRQ:
; IRQ auf INT2/PB2
; NMI auf PB1
; RDY auf PB0
;	je nach Anwendung noch ndern!!!!!
; Port C als 6510-IO-Port
.def zero = r13
;.equ ADDR_L_port = PORTA
;.equ ADDR_H_port = PORTB
;.equ DATA_port = PINC
;.equ CE_port = PORTC
.equ SYS_port = PINC		; fr IRQ, NMI und RDY
.equ CPU_DDR = DDRD			; 6502-Adresse $0000
.equ CPU_I = PIND			; 6502-Adresse $0001
.equ CPU_O = PORTD			; 6502-Adresse $0001
; Port fr NMI ist derzeit nicht aktiviert
;		=> NMI ber interne AVR-IRQ
.equ NMI_bit = 0x01
.equ IRQ_bit = 0x02
.equ RDY_bit = 0x00
.equ IRQMask = EIMSK
.equ IRQMPin = INT2

.equ ZPStart = 0x00			; Highbyte der ZP zu Beginn



;*********************************************************************************************************
;	Neu 12.11.10:
;		NEUER OPCODE: 0x0B: SZP (SetZeroPage) #$00	setzt Highbyte der ZP
;			vielleicht auch fr den Stack einfhren???
;
;
;
;			IRQ NMI RTI noch nicht getestet
;			Korrektheit der Berechnungen und Flags nicht vollstndig geprft
;				drfte eigentlich keine Probleme geben, aber wer wei
;			Noch nicht wirklich alle Wege unter allen Umstnden getestet
;				aber so strukturiert wie alles ist auch nicht ntig.
;*********************************************************************************************************

.undef XL
.undef XH


; noch frei:
;				r0	r1	r2	r3	r11
;				r21	r22	r23
;	evtl. weg	r4	r5	r9	r10


.def opcode = r20
;.def opaddr = r21
.def opbyte2 = r10
.def opbyte3 = r9
.def oparand = r8
.def ADD_L = r7
.def ADD_H = r6
.def ADC_helper1 = r5
.def ADC_helper2 = r4
.def ZP_Highbyte = r3

.def temp = r16
.def temp2 = r17
.def loop = r19

.def reg_A = r18
.def reg_X = r14
.def reg_Y = r15
.def reg_PCL = r26
.def reg_PCH = r27
.def reg_SP = r12
.def reg_Status = r25			;
.def reg_Status2 = r24
;						 7	6	5	4	3	2	1	0
;						 N	V	-	B	D	I	Z	C

;		Status2:		I	T	H	S	-	-	-	-	ORIG
;						-	-	-	B	D	I	-	-
;		Status:			-	-	-	-	V	N	Z	C


.equ flag_N = 0b00000100
.equ flag_V = 0b00001000
.equ flag_Z = 0b00000010
.equ flag_C = 0b00000001
.equ flag_I = 0b00000100
.equ flag_D = 0b00001000
.equ flag_B = 0b00010000
.equ flagN = 0x02
.equ flagV = 0x03
.equ flagZ = 0x01
.equ flagC = 0x00
.equ flagI = 0x02
.equ flagD = 0x03
.equ flagB = 0x04







.equ MHZ = 20				; ganzzahlige MHz fr die Warteschleifen
							; knnte helfen die Taktung besser hinzubekommen




; keine Ahnung, wie ich den RAM anschliee und anspreche, daher hier eigentlich nur Dummy
.macro READ_EXT_RAM
			sbic SYS_PORT, RDY_bit		; Leitung frei??
			rcall Warteschleife
			out ADDR_L_port, @0
			out ADDR_H_port, @1
			;out CE_port, 0b00000011		; RE oder so setzen
			nop; freihalten zur Sicherheit
			in @2, DATA_port

		; Variante internes SRAM:
		;	sbic IRQ_PORT, RDY_bit
		;	rcall Warteschleife
		;	mov YL, @0
		;	mov YH, @1
		;	ld @2, Y

.endmacro
;Gesamt 6 Takte!!
.macro READ_RDY
			sbic SYS_PORT, RDY_bit		; Leitung frei??
			rcall Warteschleife
.endmacro
;Gesamt 2 Takte!!

.macro WRITE_EXT_RAM
			out ADDR_L_port, @0
			out ADDR_H_port, @1
			nop
			out DATA_port, @2
.endmacro

.macro LoadZ
			ldi ZL, low (@0)
			ldi ZH, high(@0)
.endmacro



.include "WAIT_macro_inc.asm"





.org 0x00
			rjmp reset

	; alle weiteren IRQs abfangen
.org INT0addr
			rjmp NMI_Handler
.org INT1addr
			rjmp NMI_Handler
.org INT2addr
			rjmp IRQ_Handler
.org PCI0addr
			rjmp NMI_Handler
.org PCI1addr
			rjmp NMI_Handler
.org PCI2addr
			rjmp NMI_Handler
.org PCI3addr
			rjmp NMI_Handler
.org WDTaddr	
			rjmp NMI_Handler
.org OC2Aaddr
			rjmp NMI_Handler
.org OC2Baddr
			rjmp NMI_Handler
.org OVF2addr
			rjmp NMI_Handler
.org ICP1addr
			rjmp NMI_Handler
.org OC1Aaddr
			rjmp NMI_Handler
.org OC1Baddr
			rjmp NMI_Handler
.org OVF1addr
			rjmp NMI_Handler
.org OC0Aaddr
			rjmp NMI_Handler
.org OC0Baddr
			rjmp NMI_Handler
.org OVF0addr
			rjmp NMI_Handler
.org SPIaddr	
			rjmp NMI_Handler
.org URXC0addr
			rjmp NMI_Handler
.org UDRE0addr
			rjmp NMI_Handler
.org UTXC0addr
			rjmp NMI_Handler
.org ACIaddr	
			rjmp NMI_Handler
.org ADCCaddr
			rjmp NMI_Handler
.org ERDYaddr
			rjmp NMI_Handler
.org TWIaddr	
			rjmp NMI_Handler
.org SPMRaddr
			rjmp NMI_Handler


.org int_vectors_size
reset:
			ldi temp, low (RAMEND)
			out SPL, temp
			ldi temp, high(RAMEND)
			out SPH, temp

			clr zero
			ldi temp, 0x06
			mov ADC_helper1, temp
			swap temp
			mov ADC_helper2, temp

		; INT2 als IRQ konfigurieren:
			;ldi temp, 1<<INT2
			;out EIMSK, temp
		; IRQ enable:
			;sbi EIMSK, INT2




		; Daten aus dem Flash ins SRAM kopieren fr 6510er-Software
			ldi ZL, low (RAM6510_content*2)
			ldi ZH, high(RAM6510_content*2)
			ldi YL, low (RAM6510_Start)
			ldi YH, high(RAM6510_Start)
			ser temp
			;ser temp2
			mov reg_SP, temp
			ldi loop, 0x08

		reset_0:
			; einfach 4094 Bytes, egal wieviele sinnvoll sind
			lpm temp2, Z+
			st Y+, temp2
			lpm temp2, Z+
			st Y+, temp2
			dec temp
			brne reset_0
			dec loop
			brne reset_0

		; nops nur um fr den Simulator den CK auf festen 20er-Wert setzen
			nop
			nop	;hier  CK-Reset
			nop
			nop
			nop
			nop
			nop
			nop
			nop
			nop

			ldi reg_PCL, low (Vector6510Reset)			; Startadresse festlegen
			ldi reg_PCH, high(Vector6510Reset);hier fr Sim!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

			;clr ZP_Highbyte
			ldi temp, ZPStart
			mov ZP_Highbyte, temp
MAIN:
			; eigentlich noch irgendeine Synchronisation oder sowas einfgen
			rjmp main_loop			; und los :-)








.include "OP_branches.asm"
.include "OP_Stack_inc.asm"
.include "OP_System_inc.asm"
.include "OP_LD_ST_inc.asm"

.include "OP_Logical_inc.asm"





			; Interrupts verwalten		7 Takte
			; ******************
	; erstmal alle SysIRQs vom AVR als NMI ausfhren, aktivieren ber jeweiliges IO
	;		bspw. USART-IRQ
	; das I-flag hat damit aber derzeit keine Bedeutung

			; Grundstzlich:
			;	PCL + PCH auf Stack
			;	Status auf Stack
			;	neuen Vektor holen (2 byte)
			;	Sprung auf neue Adresse

	;***************************************
	; Wichtig: Rcksprung nicht ber reti sondern rjmp, spart Takte
	; evtl. ZH vorladen mit Sprungtabelle


NMI_Handler:
	; hier 3 Takte vor ld (cli, READ_RDY)
			; erstes Problem: RDY msste whrend des IRQ abgefragt werden
			sbrc reg_Status2, flagB		; Aufruf von BRK aus??
			rjmp NMIBRK_Handler

			sbr reg_Status2, flag_I
			cbi IRQMask, IRQMPin		;2

			mov YL, reg_SP	;-------1Takt
			ldi YH, 0x01
			dec YL
			RWAIT 5
			st Y, reg_PCH
			RWAIT 3
			dec YL
			st Y, reg_PCL
			;RWAIT 5		; noch Status!!
			; Status ummodeln
			mov temp, reg_Status2
			mov temp2, reg_Status
			andi temp2, 0b00000011
			or temp2, temp
			bld reg_Status, 0x02
			bst temp2, 0x07
			bld reg_Status, 0x03
			bst temp2, 0x06
			; I-Flag aus IO auslesen???
			dec YL
			mov reg_SP, YL
			RWAIT 12
			st Y, temp2		;4Takte
			RWAIT 4
			READ_RDY

			RWAIT 4
			ldi YL, low (Vector6510NMI)
			ldi YH, high(Vector6510NMI)
			READ_RDY
			ld reg_PCL, Y+
			RWAIT 4
			READ_RDY
			ld reg_PCH, Y+
			RWAIT 11
			pop temp;2
			pop temp			; AVR-Stack wieder frei machen, da kein reti
			rjmp main_loop_IRQ



IRQ_Handler:
; noch I-Flag setzen, IRQ abschalten
			sbrc reg_Status2, flagB		; Aufruf von BRK aus??
			rjmp IRQBRK_Handler

			sbr reg_Status2, flag_I
			cbi IRQMask, IRQMPin		;2

			mov YL, reg_SP	;-------1Takt
			ldi YH, 0x01
			dec YL
			RWAIT 5
			st Y, reg_PCH
			RWAIT 3
			dec YL
			st Y, reg_PCL
			;RWAIT 5		; noch Status!!
			; Status ummodeln
			mov temp, reg_Status2
			mov temp2, reg_Status
			andi temp2, 0b00000011
			or temp2, temp
			bld reg_Status, 0x02
			bst temp2, 0x07
			bld reg_Status, 0x03
			bst temp2, 0x06
			; I-Flag aus IO auslesen???
			dec YL
			mov reg_SP, YL
			RWAIT 12
			st Y, temp2		;4Takte
			RWAIT 4
			READ_RDY

			RWAIT 4
			ldi YL, low (Vector6510IRQ)
			ldi YH, high(Vector6510IRQ)
			READ_RDY
			ld reg_PCL, Y+
			RWAIT 4
			READ_RDY
			ld reg_PCH, Y+
			RWAIT 11
			pop temp
			pop temp			; AVR-Stack wieder frei machen, da kein reti
			rjmp main_loop_IRQ


IRQBRK_Handler:
		; PC und Status schon auf dem Stack
		; hier 0-Linie, ab hier 3 Takte
			ldi YL, low (Vector6510IRQ)
			ldi YH, high(Vector6510IRQ)
			RWAIT 4
			READ_RDY
			ld reg_PCL, Y+
			RWAIT 4
			READ_RDY
			ld reg_PCH, Y+
			RWAIT 11
			pop temp
			pop temp
			rjmp main_loop_IRQ


NMIBRK_Handler:
		; PC und Status schon auf dem Stack
			ldi YL, low (Vector6510NMI)
			ldi YH, high(Vector6510NMI)
			RWAIT 4
			READ_RDY
			ld reg_PCL, Y+
			RWAIT 4
			READ_RDY
			ld reg_PCH, Y+
			RWAIT 11
			pop temp;2
			pop temp
			rjmp main_loop_IRQ







main_loop:
			sei
			nop
main_loop_IRQ:
			cli
			; *************************
			; opcode lesen
			READ_RDY
			ld opcode, X+

opcode_switch:
			ldi ZL, low (opcodes_table)
			ldi ZH, high(opcodes_table)
			nop
			add ZL, opcode
			adc ZH, zero
			nop
			nop

			ijmp			; +rjmp = 4 Takte



Warteschleife:
; wenn RDY zeigt, dass Leitung nicht frei
; hier warten, bis wieder frei
	; sbic RDY
	; rcall 4
	; READ
			RWAIT 7		;13
			sbis SYS_PORT, RDY_bit
			ret		;5
			RWAIT 17		;7
			rjmp Warteschleife
		; Fr andere als m644: ret 1 Takt weniger!!!!!!!!!!!!!!!!!!!!!








.include "OPcodeTabelle.asm"
.include "RW_switch.asm"

.include "OP_IN_DE_inc.asm"
.include "OP_CP_inc.asm"
.include "OP_Shift_inc.asm"
.include "OP_illegal.asm"
.include "OP_ADC_SBC_inc.asm"

.include "6510CodeSeg_inc.asm"
.include "SRAM_inc.asm"
.include "notizen_inc.asm"

























